from abc import ABCMeta, abstractproperty
import warnings
import numpy as np

__doc__ = r"""

# Working with homogenous coordinates

The vector used to represent 2D and 3D points are vertical vectors, which are stored as 2D matrices in `numpy`.
Furthemore, in homogenous coordinates: a 3D point (x,y,z) in the world is represented by a 4 element vector
(𝜆x,𝜆y,𝜆z,𝜆) where 𝜆 ∈ ℝ₀.

To simplify access to x and y (and z) coordinates of those points as well as computations in homogenous coordinates,
we defined the types [`Point2D`](#Point2D) (and [`Point3D`](#Point3D))
extending `numpy.ndarray`. Therefore, access to y coordinate of point is `point.y` instead of
`point[1][0]` (`point[1][:]` for an array of points), and access to homogenous coordinates is made easy with `point.H`,
while it is still possible to use point with any numpy operators.

"""


class HomogeneousCoordinatesPoint(np.ndarray, metaclass=ABCMeta):
    """ Extension of Numpy `np.ndarray` that implements generic homogenous coordinates points
        for `Point2D` and `Point3D` objects. The constructor supports multiple formats for creation
        of a single point or array of multiple points.

        Example with creation of `Point2D` objects, all formulations are equivalent:
        ```
        >>> x, y = 1, 2
        >>> Point2D(x, y)
        >>> Point2D(4*x, 4*y, 4)
        >>> Point2D(np.array([[x], [y]]))
        >>> Point2D(4*np.array([[x], [y], [1]]))
        Point2D([[1.],
                 [2.]])
        ```
    """
    def __new__(cls, *coords):
        if len(coords) == 1:
            if isinstance(coords, list):
                coords = np.hstack(coords)
            else:
                coords = coords[0]
        array = np.array(coords) if isinstance(coords, (tuple, list)) else coords
        invalid_shape_message = "Invalid input shape:\n" \
            "Expected a 2D np.array of shape ({l1},N) or (N,{l1},1) in non-homogenous coordinates\n" \
            "                       or shape ({l2},N) or (N,{l2},1) in homogenous coordinates\n" \
            "Received a np.array of shape {shape}".format(l1=cls.D, l2=cls.D+1, shape=array.shape)
        if len(array.shape) == 1:
            array = array[:, np.newaxis]
        elif len(array.shape) == 2:
            pass
        elif len(array.shape) == 3:
            array = array[..., 0].T
        else:
            raise ValueError(invalid_shape_message)

        if array.shape[0] == cls.D: # point(s) given in non-homogenous coordinates
            pass
        elif array.shape[0] == cls.D+1: # point(s) given in homogenous coordinates
            array = array[0:cls.D,:]/array[cls.D,:]
        elif array.shape[0] == 0: # point given from an empty list should be an empty point
            array = np.empty((cls.D,0))
        else:
            raise ValueError(invalid_shape_message)
        return array.astype(np.float64).view(cls)
    # def __array_ufunc__():
    #    TODO
    # def __array_wrap__(self, out_arr, context=None):
    #     return super().__array_wrap__(self, out_arr, context)
    @property
    @abstractproperty
    def _coord_names(self):
        raise NotImplementedError

    x = property(fget=lambda self: self._get_coord(0), fset=lambda self, value: self._set_coord(0, value), doc="Point's x component")
    y = property(fget=lambda self: self._get_coord(1), fset=lambda self, value: self._set_coord(1, value), doc="Point's y component")
    z = property(fget=lambda self: self._get_coord(2), fset=lambda self, value: self._set_coord(2, value), doc="Point's z component (only valid for `Point3D` objects)")

    @property
    def H(self):
        """ Point expressed in homogenous coordinates with an homogenous component equal to `1`.

        Example:
        ```
        >>> p = Point3D(1,2,3)
        >>> p.H
        array([[1.],
               [2.],
               [3.],
               [1.]])
        ```
        """
        return np.vstack((self, np.ones((1, self.shape[1]))))
    # @property
    # def D(self):
    #     """ Returns the number of spacial dimensions """
    #     return len(self._coord_names)
    # /!\ a different getitem gives too much trouble with all numpy operator
    #def __getitem__(self, i):
    #    if isinstance(i, int):
    #        return self.__class__(super().__getitem__((slice(None), i)))
    #    return super().__getitem__(i)

    # /!\ iter may conflict with numpy array getitem.
    def __iter__(self):
        return (self.__class__(self[:,i:i+1]) for i in range(self.shape[1]))

    def to_list(self):
        """ Transforms a single point to a python list.

        Raises:
            AssertionError if the object is an array of multiple points
        """
        assert self.shape[1] == 1, "to_list() method can only be used on single point {}".format(self.__class__.__name__)
        return self[:,0].flatten().tolist()

    def flatten(self):
        """ Flatten the points.

        .. todo:: integrate this in the __array_ufunc__ to prevent type forwarding
        """
        return np.asarray(super().flatten())

    def to_int_tuple(self):
        """ Transforms a single point to a python tuple with integer coordinates
        """
        return tuple(int(x) for x in self.to_list())

    def linspace(self, num):
        """ Linearly interpolate points in `num-1` intervals.

            Example:
            ```
            >>> Point2D([0,4,4],[0,0,4]).linspace(5)
            [[0. 1. 2. 3. 4. 4. 4. 4. 4. 4.]
            [0. 0. 0. 0. 0. 0. 1. 2. 3. 4.]]
            ```
        """
        return np.transpose(np.linspace(self[:,:-1], self[:,1:], num), axes=(1,2,0)).reshape(len(self._coord_names),-1)

    def close(self):
        """ Copy the first point in an array of points and place it at the end of that array,
            hence "closing" the polygon defined by the initial points.

            TODO: add Points2D and Points3D classes
        """
        assert self.shape[1] > 1, f"Invalid use of 'close' method: points' shape '{self.shape}' expected to be > 1 in the second dimension"
        return self.__class__(np.hstack((self, self[:,0:1])))

    _get_coord = lambda self, i:        np.asarray(super().__getitem__((i,0)))       if self.shape[1] == 1 else np.asarray(super().__getitem__(i))
    _set_coord = lambda self, i, value:            super().__setitem__((i,0), value) if self.shape[1] == 1 else            super().__setitem__((i), value)


class Point2D(HomogeneousCoordinatesPoint):
    """ Numpy representation of a single 2D point or a list of 2D points
    """
    D = 2
    _coord_names = ("x","y")

class Point3D(HomogeneousCoordinatesPoint):
    """ Numpy representation of a single 3D point or a list of 3D points
    """
    D = 3
    _coord_names = ("x","y","z")
    @property
    def V(self):
        array = self.H
        array[-1] = 0
        return VanishingPoint(array)

class VanishingPoint(Point3D):
    """ Object allowing representation of Vanishing point (with null homogenous
        coordinate). Only the `H` attribute should be used. Handle with care.
    """
    def __new__(cls, array):
        warnings.warn("Vanishing Point feature has not yet been fully tested")
        obj = array.astype(np.float64).view(cls)
        obj.array = array
        return obj
    @property
    def H(self):
        return self.array
    def __getattribute__(self, attr_name):
        if attr_name not in ("H", "__array_finalize__", "array", "shape", "size", "ndim", "x", "y", "z", "_get_coord", "close", "__class__", "astype", "view"):
            raise AttributeError(f"VanishingPoint has no `{attr_name}` attribute.")
        return super().__getattribute__(attr_name)



import cv2
import tensorflow as tf
from calib3d import Calib
# pylint: disable=unexpected-keyword-arg, no-value-for-parameter, too-many-function-args

class TensorflowCalib():
    def __init__(self, *, width, height, T=None, K, kc, r=None, R=None, Kinv=None, Pinv=None, P=None, dtype=tf.float32):
        self.width = tf.cast(width, dtype=dtype)
        self.height = tf.cast(height, dtype=dtype)
        self.K = tf.cast(K, dtype=dtype)
        self.T = tf.cast(T, dtype=dtype)
        self.r = tf.cast(r, dtype=dtype)
        self.R = tf.cast(R, dtype=dtype) if R is not None else rodrigues_batch(r, dtype=dtype)

        self.P = tf.cast(P, dtype=dtype) if P is not None else tf.matmul(self.K, tf.concat((self.R, self.T), axis=-1))
        self.Pinv = tf.cast(Pinv, dtype=dtype) if Pinv is not None else pinv(self.P, dtype=dtype)
        self.Kinv = tf.cast(Kinv, dtype=dtype) if Kinv is not None else pinv(self.K, dtype=dtype)

        self.kc = tf.cast(kc, dtype=dtype) if kc is not None else None
        self.C = -tf.matmul(self.R, self.T, transpose_a=True) # pylint: disable=invalid-unary-operand-type

        self.batch_size = K.shape[0]
        self.dtype = dtype
    @classmethod
    def from_numpy(cls, calib: Calib, dtype=tf.float64):
        return cls(
            K=tf.constant(calib.K, dtype=dtype)[tf.newaxis],
            r=tf.constant(cv2.Rodrigues(calib.R)[0], dtype=dtype)[tf.newaxis],
            T=tf.constant(calib.T, dtype=dtype)[tf.newaxis],
            width=tf.constant(calib.width, dtype=dtype)[tf.newaxis],
            height=tf.constant(calib.height, dtype=dtype)[tf.newaxis],
            kc=tf.constant(calib.kc, dtype=dtype)[tf.newaxis],
            dtype=dtype
        )
    def project_3D_to_2D(self, point3D):
        point2D = from_homogenous(tf.matmul(self.P, to_homogenous(point3D, dtype=self.dtype)))
        # TODO: avoid distort points too much outside the image as the distortion model is not perfect
        return self.distort(point2D)
    def project_2D_to_3D(self, point2D, Z):
        point2D = self.rectify(point2D)
        X = from_homogenous(tf.matmul(self.Pinv, to_homogenous(point2D, dtype=self.dtype)))
        d = (X - self.C)
        P = batch_expand(tf.constant([[0],[0],[Z]], dtype=self.dtype), d)
        n = batch_expand(tf.constant([[0],[0],[1]], dtype=self.dtype), d)
        return find_intersection(self.C, d, P, n)
    def distort(self, point2D):
        if self.kc is None:
            return point2D
        point2D = from_homogenous(tf.matmul(self.Kinv, to_homogenous(point2D, dtype=self.dtype)))
        rad1, rad2, tan1, rad3, tan2 = self.kc[:,0], self.kc[:,1], self.kc[:,2], self.kc[:,3], self.kc[:,4]
        r2 = point2D[:,0]*point2D[:,0] + point2D[:,1]*point2D[:,1]
        delta = 1 + rad1[:,tf.newaxis]*r2 + rad2[:,tf.newaxis]*r2*r2 + rad3[:,tf.newaxis]*r2*r2*r2
        dx = tf.stack((
            2*tan1[:,tf.newaxis]*point2D[:,0]*point2D[:,1] + tan2[:,tf.newaxis]*(r2 + 2*point2D[:,0]*point2D[:,0]),
            2*tan2[:,tf.newaxis]*point2D[:,0]*point2D[:,1] + tan1[:,tf.newaxis]*(r2 + 2*point2D[:,1]*point2D[:,1])
        ), axis=1)
        point2D = point2D*delta + dx
        return from_homogenous(tf.matmul(self.K, to_homogenous(point2D, dtype=self.dtype)))
    def rectify(self, point2D):
        if self.kc is None:
            return point2D
        point2D = from_homogenous(tf.matmul(self.Kinv, to_homogenous(point2D, dtype=self.dtype)))
        rad1, rad2, tan1, rad3, tan2 = self.kc[:,0], self.kc[:,1], self.kc[:,2], self.kc[:,3], self.kc[:,4]
        r2 = point2D[:,0]*point2D[:,0] + point2D[:,1]*point2D[:,1]
        delta = 1 + rad1[:,tf.newaxis]*r2 + rad2[:,tf.newaxis]*r2*r2 + rad3[:,tf.newaxis]*r2*r2*r2
        dx = tf.stack((
            2*tan1[:,tf.newaxis]*point2D[:,0]*point2D[:,1] + tan2[:,tf.newaxis]*(r2 + 2*point2D[:,0]*point2D[:,0]),
            2*tan2[:,tf.newaxis]*point2D[:,0]*point2D[:,1] + tan1[:,tf.newaxis]*(r2 + 2*point2D[:,1]*point2D[:,1])
        ), axis=1)
        point2D = (point2D - dx)/delta[:,tf.newaxis]
        return from_homogenous(tf.matmul(self.K, to_homogenous(point2D, dtype=self.dtype)))

def batch_expand(input_tensor, batch_tensor):
    # https://stackoverflow.com/questions/57716363/explicit-broadcasting-of-variable-batch-size-tensor
    length = len(batch_tensor.shape)-1
    input_tensor = input_tensor[tf.newaxis]
    broadcast_shape = tf.where([True, *[False]*length], tf.shape(batch_tensor), tf.shape(input_tensor))
    return tf.broadcast_to(input_tensor, broadcast_shape)

def to_homogenous(points, dtype=tf.float64):
    _,_,N = points.shape
    ones = tf.ones((1,N), dtype=dtype)
    ones = batch_expand(ones, points)
    return tf.concat((points, ones), axis=-2)

def from_homogenous(points):
    return points[:,:-1,:]/points[:,-1:,:]

def pinv(a, rcond=1e-15, dtype=tf.float64):
    s, u, v = tf.svd(a)
    # Ignore singular values close to zero to prevent numerical overflow
    limit = rcond * tf.reduce_max(s)
    non_zero = tf.greater(s, limit)

    reciprocal = tf.where(non_zero, tf.cast(tf.reciprocal(s), dtype=dtype), tf.zeros_like(s))
    lhs = tf.matmul(v, tf.matrix_diag(reciprocal))
    return tf.matmul(lhs, u, transpose_b=True)

def find_intersection(C, d, P, n):
    d = d/tf.norm(d, axis=-2, keepdims=True)
    dist = tf.tensordot(P-C, n, axes=[[1],[1]])[:,:,0] / tf.tensordot(d, n, axes=[[1],[1]])[:,tf.newaxis,:,0,0]  # Distance between plane z=Z and camera
    return C + dist*d

# https://github.com/blzq/tf_rodrigues/blob/master/rodrigues.py
def rodrigues_batch(rvecs, dtype=tf.float64):
    """ Convert a batch of axis-angle rotations in rotation vector form shaped
        (batch, 3) to a batch of rotation matrices shaped (batch, 3, 3).
        See
        https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula#Matrix_notation
        https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    """
    rvecs = tf.cast(rvecs, dtype=dtype)
    batch_size = tf.shape(rvecs)[0]
    assert rvecs.shape[1] == 3

    thetas = tf.norm(rvecs, axis=1, keepdims=True)
    is_zero = tf.equal(tf.squeeze(thetas), 0.0)
    u = rvecs / thetas
    # Each K is the cross product matrix of unit axis vectors
    # pyformat: disable
    zero = tf.zeros([batch_size], dtype=dtype)  # for broadcasting
    Ks_1 = tf.stack([  zero   , -u[:, 2],  u[:, 1] ], axis=1)  # row 1
    Ks_2 = tf.stack([  u[:, 2],  zero   , -u[:, 0] ], axis=1)  # row 2
    Ks_3 = tf.stack([ -u[:, 1],  u[:, 0],  zero    ], axis=1)  # row 3
    # pyformat: enable
    Ks = tf.stack([Ks_1, Ks_2, Ks_3], axis=1)                  # stack rows

    Rs = tf.eye(3, batch_shape=[batch_size], dtype=dtype) + \
         tf.sin(thetas)[..., tf.newaxis] * Ks + \
         (1 - tf.cos(thetas)[..., tf.newaxis]) * tf.matmul(Ks, Ks)

    # Avoid returning NaNs where division by zero happened
    return tf.where(is_zero, tf.eye(3, batch_shape=[batch_size], dtype=dtype), Rs)



__doc__ = r"""
This library offers several tools for manipulation of calibrated cameras, projective geometry and computations using homogenous coordinates.

Camera calibration allows to determine the relation between the camera's pixels (2D coordinates) and points in the real world
(3D coordinates). It implies computation using homogenous coordinates. This python library aims at simplifying implementations
of projective geometry computations, building on top of `numpy` and `cv2`.

The different modules are document here:

- [Computations with homogenous coordinates](./points.html)
- [Projective geometry and calibrated cameras](./calib.html)
- [Drawing 3D objects on a 2D canvas using projective geometry](./draw.html)

"""

from .points import Point3D, Point2D, VanishingPoint
from .calib import Calib, parameters_to_affine_transform, compute_rotation_matrix, line_plane_intersection, compute_shear_rectification_matrix
from .draw import ProjectiveDrawer



import cv2
import numpy as np

from calib3d import Calib, Point2D, Point3D

__doc__ = r"""

# Projective Drawing

When drawing 3D objets on a 2D canvas, several things need to be considered:
- Projection onto the 2D space using the calibration information
- Handling lens distortion that make straight lines appear curved
- Handling of objects visiblity given the canvas dimensions
"""

class ProjectiveDrawer():
    """ Given the calibration information with `Calib`, and a number of segments
        to decompose straight lines, this objet offer several functions to draw
        on a 2D canvas given 3D coordinates.
    """
    def __init__(self, calib: Calib, color, thickness: int=1, segments: int=10):
        self.color = color
        self.thickness = thickness
        self.calib = calib
        self.segments = segments

    def _polylines(self, canvas, points: Point2D, color=None, thickness: int=None, markersize=None, **kwargs):
        thickness = thickness or self.thickness
        color = color or self.color
        if isinstance(canvas, np.ndarray):
            points = np.array(points.astype(np.int32).T.reshape((-1,1,2)))
            if thickness < 0:
                cv2.fillPoly(canvas, [points], color=color, **kwargs)
            else:
                cv2.polylines(canvas, [points], False, color=color, thickness=thickness, **kwargs)
            if markersize:
                for point in points:
                    cv2.drawMarker(canvas, tuple(point[0]), color=color, markerSize=markersize, thickness=thickness)
        else:
            if thickness < 0:
                try:
                    import matplotlib as mpl
                    p = mpl.patches.Polygon(np.array(list(zip(points.x, points.y))), facecolor=np.array(color)/255, alpha=.35)
                    canvas.add_patch(p)
                except ImportError:
                    raise ImportError("The current implementation requires matplotlib")
            else:
                canvas.plot(points.x, points.y, linewidth=thickness, color=np.array(color)/255, markersize=markersize, **kwargs)

    def polylines(self, canvas, points3D: Point3D, color=None, thickness: int=None, **kwargs):
        self._polylines(canvas, self.calib.project_3D_to_2D(points3D), color=color, thickness=thickness, **kwargs)
        #for point3D1, point3D2 in zip(points3D, points3D.close()[:,1:]):
        #    self.draw_line(canvas, point3D1, point3D2, *args, **kwargs)

    def draw_line(self, canvas, point3D1: Point3D, point3D2: Point3D, color=None, thickness: int=None, only_visible=True, **kwargs):
        if only_visible:
            try:
                point3D1, point3D2 = visible_segment(self.calib, point3D1, point3D2)
            except ValueError:
                return
        points3D = Point3D(np.linspace(point3D1, point3D2, self.segments+1))
        self._polylines(canvas, self.calib.project_3D_to_2D(points3D), color=color, thickness=thickness, **kwargs)

    def draw_arc(self, canvas, center, radius, start_angle=0.0, stop_angle=2*np.pi, color=None, thickness=None):
        thickness = thickness or self.thickness
        angles = np.linspace(start_angle, stop_angle, self.segments*4+1)
        xs = np.cos(angles)*radius + center.x
        ys = np.sin(angles)*radius + center.y
        zs = np.ones_like(angles)*center.z
        points3D = Point3D(np.vstack((xs,ys,zs)))
        self._polylines(canvas, self.calib.project_3D_to_2D(points3D), color=color, thickness=thickness)

    def draw_rectangle(self, canvas, point3D1, point3D2):
        c1 = point3D1
        c3 = point3D2
        if point3D1.z == point3D2.z:
            c2 = Point3D(c1.x, c3.y, c1.z)
            c4 = Point3D(c3.x, c1.y, c1.z)
        elif point3D1.x == point3D2.x:
            c2 = Point3D(c1.x, c1.y, c3.z)
            c4 = Point3D(c1.x, c3.y, c1.z)
        elif point3D1.y == point3D2.y:
            c2 = Point3D(c1.x, c1.y, c3.z)
            c4 = Point3D(c3.x, c1.y, c1.z)
        corners = [c1, c2, c3, c4, c1]
        for p1, p2 in zip(corners, corners[1:]):
            self.draw_line(canvas, p1, p2)

    def fill_polygon(self, canvas, points3D):
        points3D = points3D.close().linspace(self.segments)
        self._polylines(canvas, self.calib.project_3D_to_2D(points3D), thickness=-1)

def visible_segment(calib: Calib, point3D1: Point3D, point3D2: Point3D):
    """ From a segment defined by the given two 3D points, compute the two 3D
        points delimiting the visible segment in the given calib.
    """
    def dichotomy(inside, outside, max_it=10):
        middle = Point3D((inside+outside)/2)
        if max_it == 0:
            return middle
        max_it = max_it - 1
        return dichotomy(middle, outside, max_it) if calib.projects_in(middle) else dichotomy(inside, middle, max_it)
    def find_point_inside(p1, p2, max_it=4):
        assert not calib.projects_in(p1) and not calib.projects_in(p2)
        middle = Point3D((p1+p2)/2)
        if calib.projects_in(middle):
            return middle
        if max_it == 0:
            return None
        point_inside = find_point_inside(middle, p2, max_it-1)
        if point_inside is not None:
            return point_inside
        return find_point_inside(middle, p1, max_it-1)

    p1, p2 = point3D1, point3D2
    if calib.projects_in(p1) and calib.projects_in(p2):
        return p1, p2
    elif calib.projects_in(p1):
        return p1, dichotomy(p1, p2)
    elif calib.projects_in(p2):
        return dichotomy(p2, p1), p2
    else:
        point_inside = find_point_inside(p1, p2)
        if point_inside is None:
            raise ValueError
        return dichotomy(point_inside, p1), dichotomy(point_inside, p2)



import pickle
import numpy as np
import cv2
from .points import Point2D, Point3D, HomogeneousCoordinatesPoint

__doc__ = r"""

# Theoretical introduction

Camera calibration is based on the pinhole camera model that approximate how lights travels between the scene and the camera sensor. There are two sets of parameters:

- The **extrinsic parameters** define the camera position and orientation relative to the world coordinates system.
- The **Intrinsic parameters** define the camera sensor and lens parameters.

There are 3 different coordinates systems:

- The world 3D coordinates system
- The camera 3D coordinates system
- The 2D pixel positions where 3D positions are being projected.


## Extrinsic parameters
The extrinsic parameters defines the transformations from the world 3D coordinates \(\left[x_O,y_O,z_O\right]^T\) to the camera 3D coordinates \(\left[x_C,y_C,z_C\right]^T\). The camera 3D coordinates system has the following **conventions**:

- The point \((0,0,0)\) is the center of projection of the camera and is called the *principal point*.
- The \(z\) axis of the camera points *towards the scene*, the \(x\) axis is along the sensor width pointing towards the right, and the \(y\) axis is along the sensor height pointing towards the bottom.

The camera coordinates system is therefore a *transformation* of the world coordinates systems with:

- A **rotation** defined by a rotation matrix \(R\) using euler angles in a right-hand orthogonal system. The rotation is applied to the world coordinates system to obtain the camera orientation.
- A **translation** defined by a translation vector \(T\) representing the position of the center of the world in the camera coordinates system !

Hence,

$$\lambda\left[\begin{matrix}x_C\\ y_C\\ z_C\\ 1\end{matrix}\right] = \left[\begin{matrix}
R_{3\times 3} & T_{3\times 1}\\{\bf 0}_{1\times 3}&1
\end{matrix}\right] \left[\begin{matrix}x_O\\y_O\\z_O\\1\end{matrix}\right]$$

**Important notes:**

- The rotation matrix represents a passive (or alias) transformation because it's the coordinates system that rotates and not the objects.
- Euler angles define a 3D rotation starting with a rotation around \(x\) followed by a rotation around \(y\) followed by a rotation around \(z\) (the order matters).
- If \(T\) is expressed in the camera coordinate system, the position of the camera expressed in world coordinates system is \(C:=-R^{-1}T = -R^{T}T\) (since \(R\) is a rotation matrix).


## Intrinsic parameters

The intrinsic parameters defines the transformation between the 3D coordinates relative to the camera center \(\left[x_C,y_C,z_C\right]^T\) and the 2D coordinates in the camera sensor \(\left[i,j\right]^T\). This transformation is called a *projection* and includes:

- the scale produced by the focal length, with \(f\) being the distance between the camera center and the plane on which the image is projected.
- the scale factors \((m_x,m_y)\) relating pixels units to distance units (usually \(m_x=m_y\) because pixels are squares).
- the translation from the camera _principal point_ to a top-left origin, with \((u_0,v_0)\) being the position of the *principal point* expressed in the image coordinates system.
- a skew coefficient \(\gamma\) between the \(x\) and \(y\) axis in the sensor (usually \(\gamma=0\) because pixels are squares).

Those transformations can be aggregated in one single matrix called **camera matrix**:

$$K := \left[\begin{matrix}f\cdot m_x & \gamma & u_0 \\ 0 & f\cdot m_y & v_0 \\ 0 & 0 & 1\end{matrix}\right]$$

Therefore,
$$\lambda\left[\begin{matrix}i\\ j\\ 1\end{matrix}\right]= \left[\begin{matrix}K_{3\times 3}&{\bf 0}_{3\times 1}\end{matrix}\right]\left[\begin{matrix}x_C\\y_C\\z_C\\1\end{matrix}\right]$$

**Notes:**

- The width and height of the image are to be added to those parameters and delimits the sensor width and height in pixels.
- When applying the **direct** projection of a given 3D point, different values of \(\lambda\) will always give the **same** 2D point.
- When applying the **inverse** projection on a given 2D point, different values of \(\lambda\) will give **different** 3D points.

This is obvious when simplifying the relation between the two points (The column \({\bf 0}_{3\times 1}\) cancels the homogenous component of the 3D point):

$$\lambda\left[\begin{matrix}i\\j\\1\end{matrix}\right]= \left[\begin{matrix}K_{3\times 3}\end{matrix}\right]\left[\begin{matrix}x_C\\y_C\\z_C\end{matrix}\right]$$

The 2D vector in homogenous coordinates is not affected by the value of \(\lambda\), while the 3D vector is.



## Projection model

Therefore, by combining

- the transformation from the world coordinates system to the camera coordinates system (defined by \(R\) and \(T\))
- with the projection from the camera coordinates system to the image pixels (defined by \(K\)),

We have a projection model allowing to compute the coordinates of a 2D point in the image \(\left(i,j\right)\) from a 3D point in the real world \(\left(x,y,z\right)\) described by the matrix \(P\):
$$P := \left[\begin{matrix}K_{3\times 3}&{\bf 0}_{3\times 1}\end{matrix}\right] \left[\begin{matrix}R_{3\times 3}&T_{3\times 1}\\{\bf 0}_{1\times 3}&1\end{matrix}\right]=K_{3\times 3}\left[\begin{matrix}R_{3\times 3}&T_{3\times 1}\end{matrix}\right]$$

The opposite operation requires to invert \(P\) and is done by pseudo-inverse inversion because \(P\) is rectangular.


## Limitations
We need a model of the distortions brought by the lens:

- **radial** distortion cause lines far from principal point to look distorted.
- **tangential** distortion: occur when lens is not prefectly align with the \(z\) axis of the camera.

Many lens distortion models exist. Here, we will use the model used in [opencv](https://docs.opencv.org/3.1.0/d4/d94/tutorial_camera_calibration.html).


## Distortion models

There are actually 2 different models:

- The direct model that applies distortion: find the 2D image (distorted) coordinates of a 3D point given its 2D projected (undistorted) coordinates.
- The inverse model that rectifies distortion: find the 2D (undistorted) coordinate allowing to project a 2D (distorted) image coordinate into 3D.

.. image:: https://github.com/ispgroupucl/calib3d/blob/main/assets/distortion_steps.png?raw=true

### Direct model: "distort"

The following relationships allows to compute the distorted 2D coordinates \((x_d,y_d)\) on an image from the 2D coordinates provided by the linear projection model \((x_u,y_u)\). Those coordinates are expressed in the camera coordinate system, i.e. they are distances (not pixels) and the point \((0,0)\) is the principal point of the camera.

$$\begin{align}
    x_d &= x_u\overbrace{\left(1 + k_1 {r_u}^2 + k_2 {r_u}^4 + k_3 {r_u}^6+\cdots\right)}^{\text{radial component}} + \overbrace{\left[2p_1 x_uy_u + p_2\left({r_u}^2 + 2{x_u}^2\right)\right]}^{\text{tangeantial component}}\\
    y_d &= y_u\left(1 + k_1 {r_u}^2 + k_2 {r_u}^4 + k_3 {r_u}^6+\cdots\right) + \left[2p_2 x_uy_u + p_1\left({r_u}^2 + 2{y_u}^2\right)\right]
\end{align}$$

Where:

- \(k_1, k_2, k_3, \cdots\)  are the radial distortion coefficients
- \(t_1, t_2\) are the tangential distortion coefficients
- \({r_u}^2 := {x_u}^2 + {y_u}^2\)

We usually use only 3 radial distortion coefficients, which makes a total of 5 coefficients. Those coefficients are found by running an optimisation algorithm on a set of 2D point - 3D point relations as we did with `cv2.calibrateCamera`. They are stored in the `k` vector.

### Inverse model: "rectify"

The distortion operation cannot be inverted analitically using the coefficients \(k_1\), \(k_2\), \(k_3\), \(p_1\), \(p_2\) (i.e. have \(x_u=f_{k_{1,2,3},p_{1,2}}(x_d,y_d)\) and \(y_u=f_{k_{1,2,3},p_{1,2}}(x_d,y_d)\)). We either need another model with another set of coefficients, or make an approximation.

Here, we will use the following approximation: We will assume that the distortion at point \((x_d,y_d)\) would be the same that distortion at point \((x_u,y_u)\) ! Therefore:

$$\left\{\begin{align}
    2p_1 x_uy_u + p_2\left({r_u}^2 + 2{x_u}^2\right) &\approx 2p_1 x_dy_d + p_2\left({r_d}^2 + 2{x_d}^2\right)\\
    2p_2 x_uy_u + p_1\left({r_u}^2 + 2{y_u}^2\right) &\approx 2p_2 x_dy_d + p_1\left({r_d}^2 + 2{y_d}^2\right) \\
    \left(1 + k_1 {r_u}^2 + k_2 {r_u}^4 + k_3 {r_u}^6+\cdots\right) &\approx \left(1 + k_1 {r_d}^2 + k_2 {r_d}^4 + k_3 {r_d}^6+\cdots\right)
   \end{align}\right.
    $$

If this approximation holds, it's much easier to get an analytical expression of \(x_u=f_{k_{1,2,3},p_{1,2}}(x_d,y_d)\) and \(y_u=f_{k_{1,2,3},p_{1,2}}(x_d,y_d)\).


# Implementation

This library defines a [Calib](./calib3d.calib#Calib) object to represent a calibrated camera. Its constructor receives in arguments the intrinsic and extrinsic parameters:

- image dimensions `width` and `height`,
- the translation vector `T`,
- the rotation matrix `R`,
- the camera matrix `K`.

The method `project_3D_to_2D` allows to compute the position in the image of a 3D point in the world. The opposite operation `project_2D_to_3D` requires an additional parameter `Z` that tells the \(z\) coordinate of the 3D point.

"""

EPS = 1e-5

class Calib():
    def __init__(self, *, width: int, height: int, T: np.ndarray, R: np.ndarray, K: np.ndarray, k=None, **_) -> None:
        """ Represents a calibrated camera.

        Args:
            width (int): image width
            height (int): image height
            T (np.ndarray): translation vector
            R (np.ndarray): rotation matrix
            K (np.ndarray): camera matrix holding intrinsic parameters
            k (np.ndarray, optional): lens distortion coefficients. Defaults to None.
        """
        self.w = self.width = int(width)
        self.h = self.height = int(height)
        self.T = T
        self.K = K
        self.k = np.array((k if k is not None else [0,0,0,0,0]), dtype=np.float64)
        self.R = R
        self.C = Point3D(-R.T@T)
        self.P = self.K @ np.hstack((self.R, self.T))
        self.Pinv = np.linalg.pinv(self.P)
        self.Kinv = np.linalg.pinv(self.K)

    def update(self, **kwargs) -> 'Calib':
        """ Creates another Calib object with the given keyword arguments updated
            Args:
                **kwargs : Any of the arguments of `Calib`. Other arguments remain unchanged.
            Returns:
                A new Calib object
        """
        return self.__class__(**{**self.dict, **kwargs})

    @classmethod
    def from_P(cls, P, width, height) -> 'Calib':
        """ Create a `Calib` object from a given projection matrix `P` and image dimensions `width` and `height`.

            Args:
                P (np.ndarray): a 3x4 projection matrix
                width (int): image width
                height (int): image height
            Returns:
                A Calib object
        """
        K, R, T, Rx, Ry, Rz, angles = cv2.decomposeProjectionMatrix(P) # pylint: disable=unused-variable
        return cls(K=K, R=R, T=Point3D(-R@Point3D(T)), width=width, height=height)

    @classmethod
    def load(cls, filename) -> 'Calib':
        """ Loads a Calib object from a file (using the pickle library)
            Args:
                filename (str): the file that stores the Calib object
            Returns:
                The `Calib` object previously saved in `filename`.
        """
        with open(filename, "rb") as f:
            return cls(**pickle.load(f))

    @property
    def dict(self) -> dict:
        """ Gets a dictionnary representing the calib object (allowing easier serialization)
        """
        return {k: getattr(self, k) for k in self.__dict__}

    def dump(self, filename) -> None:
        """ Saves the current calib object to a file (using the pickle library)
            Args:
                filename (str): the file that will store the calib object
        """
        with open(filename, "wb") as f:
            pickle.dump(self.dict, f)

    def _project_3D_to_2D_cv2(self, point3D: Point3D):
        raise BaseException("This function gives errors when rotating the calibration...")
        return Point2D(cv2.projectPoints(point3D, cv2.Rodrigues(self.R)[0], self.T, self.K, self.k.astype(np.float64))[0][:,0,:].T)

    def project_3D_to_2D(self, point3D: Point3D) -> Point2D:
        """ Using the calib object, project a 3D point in the 2D image space.
            Args:
                point3D (Point3D): the 3D point to be projected
            Returns:
                The point in the 2D image space on which point3D is projected by calib
        """
        #assert isinstance(point3D, Point3D), "Wrong argument type '{}'. Expected {}".format(type(point3D), Point3D)
        point2D_H = self.P @ point3D.H # returns a np.ndarray object
        point2D_H[2] = point2D_H[2] * np.sign(point2D_H[2]) # correct projection of points being projected behind the camera
        point2D = Point2D(point2D_H)
        # avoid distortion of points too far away
        excluded_points = np.logical_or(np.logical_or(point2D.x < -self.width, point2D.x > 2*self.width),
                                        np.logical_or(point2D.y < -self.height, point2D.y > 2*self.height))
        return Point2D(np.where(excluded_points, point2D, self.distort(point2D)))

    def project_2D_to_3D(self, point2D: Point2D, X: float=None, Y: float=None, Z: float=None) -> Point3D:
        """ Using the calib object, project a 2D point in the 3D image space
            given one of it's 3D coordinates (X,Y or Z). One and only one
            coordinate must be given.
            Args:
                point2D (Point2D): the 2D point to be projected
                X (float): the X coordinate of the 3D point
                Y (float): the Y coordinate of the 3D point
                Z (float): the Z coordinate of the 3D point
            Returns:
                The point in the 3D world that projects on `point2D` and for
                which the given coordinates is given.
        """
        v = [X,Y,Z]
        assert sum([1 for x in v if x is None]) == 2
        assert isinstance(point2D, Point2D), "Wrong argument type '{}'. Expected {}".format(type(point2D), Point2D)
        point2D = self.rectify(point2D)
        X = Point3D(self.Pinv @ point2D.H)
        d = (X - self.C)
        P = np.nan_to_num(Point3D(*v), 0)
        v = np.array([[0 if x is None else 1 for x in v]]).T
        return line_plane_intersection(self.C, d, P, v)

    def distort(self, point2D: Point2D) -> Point2D:
        """ Applies lens distortions to the given `point2D`.
        """
        if np.any(self.k):
            rad1, rad2, tan1, tan2, rad3 = self.k.flatten()
            # Convert image coordinates to camera coordinates (with z=1 which is the projection plane)
            point2D = Point2D(self.Kinv @ point2D.H)

            r2 = point2D.x*point2D.x + point2D.y*point2D.y
            delta = 1 + rad1*r2 + rad2*r2*r2 + rad3*r2*r2*r2
            dx = np.array([
                2*tan1*point2D.x*point2D.y + tan2*(r2 + 2*point2D.x*point2D.x),
                2*tan2*point2D.x*point2D.y + tan1*(r2 + 2*point2D.y*point2D.y)
            ]).reshape(2, -1)

            point2D = point2D*delta + dx
            # Convert camera coordinates to pixel coordinates
            point2D = Point2D(self.K @ point2D.H)
        return point2D

    def rectify(self, point2D: Point2D) -> Point2D:
        """ Removes lens distortion to the given `Point2D`.
        """
        if np.any(self.k):
            rad1, rad2, tan1, tan2, rad3 = self.k.flatten()
            point2D = Point2D(self.Kinv @ point2D.H) # to camera coordinates

            r2 = point2D.x*point2D.x + point2D.y*point2D.y
            delta = 1 + rad1*r2 + rad2*r2*r2 + rad3*r2*r2*r2
            dx = np.array([
                2*tan1*point2D.x*point2D.y + tan2*(r2 + 2*point2D.x*point2D.x),
                2*tan2*point2D.x*point2D.y + tan1*(r2 + 2*point2D.y*point2D.y)
            ]).reshape(2, -1)

            point2D = (point2D - dx)/delta
            point2D = Point2D(self.K @ point2D.H) # to pixel coordinates
        return point2D

    def crop(self, x_slice: slice, y_slice: slice) -> 'Calib':
        x0 = x_slice.start
        y0 = y_slice.start
        width = x_slice.stop - x_slice.start
        height = y_slice.stop - y_slice.start
        T = np.array([[1, 0,-x0], [0, 1,-y0], [0, 0, 1]])
        return self.update(width=width, height=height, K=T@self.K)

    def scale(self, output_width: int=None, output_height: int=None, sx: float=None, sy: float=None) -> 'Calib':
        """ Returns a calib corresponding to a camera that was scaled by horizontal and vertical scale
            factors `sx` and `sy`. If `sx` and `sy` are not given, the scale factors are computed with the current
            width and height and the given `output_width` and `output_height`.
        """
        sx = sx or output_width/self.width
        sy = sy or output_height/self.height
        width = output_width or int(self.width*sx)
        height = output_height or int(self.height*sy)
        S = np.array([[sx, 0, 0], [0, sy, 0], [0, 0, 1]])
        return self.update(width=width, height=height, K=S@self.K)

    def flip(self) -> 'Calib':
        """ Returns a calib corresponding to a camera that was flipped horizontally.
        """
        F = np.array([[-1, 0, self.width-1], [0, 1, 0], [0, 0, 1]])
        return self.update(K=F@self.K)

    def rotate(self, angle) -> 'Calib':
        """ Returns a calib corresponding to a camera that was rotated by `angle` degrees.
            angle (float) : Angle of rotation in degrees.
        """
        if angle == 0:
            return self
        A, new_width, new_height = compute_rotate(self.width, self.height, angle, degrees=True)
        return self.update(K=A@self.K, width=new_width, height=new_height)

    def rot90(self, k) -> 'Calib':
        """ Returns a calib corresponding to a camera that was rotated `k` times around it's main axis.
            k (int) : Number of times the array is rotated by 90 degrees.

            .. todo:: This method should be tested.
        """
        raise NotImplementedError("Method not tested")
        R = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])**k
        transpose = k % 2
        return self.update(K=R@self.K, width=self.height if transpose else self.width, height=self.width if transpose else self.height)

    def compute_length2D(self, point3D: Point3D, length3D: float) -> np.ndarray:
        """ Returns the length in pixel of a 3D length at point3D

            .. versionchanged:: 2.8.0: `length3D` and `point3D` were interverted.
        """
        assert np.isscalar(length3D), f"This function expects a scalar `length3D` argument. Received {length3D}"
        point3D_c = Point3D(np.hstack((self.R, self.T)) @ point3D.H)  # Point3D expressed in camera coordinates system
        orth = np.dot(point3D_c.flatten(), Point3D(point3D_c.x, point3D_c.y, 0).flatten())
        orth = orth *length3D / np.linalg.norm(orth)
        point3D_c.y += orth # add the 3D length orthogonally to the point in camera coordinates
        point2D = self.distort(Point2D(self.K @ point3D_c)) # go in the 2D world
        length = np.linalg.norm(point2D - self.project_3D_to_2D(point3D), axis=0)
        return length#float(length) if point3D.shape[1] == 1 else length

    def compute_length3D(self, point3D: Point3D, length2D: float) -> np.ndarray:
        """ Returns the (maximum) distance of the 3D point located `length2D` pixels away from `point3D` in the image plane.
        """
        assert np.isscalar(length2D), f"This function expects a scalar `length2D` argument. Received {length2D}"
        point2D_shifted = self.project_3D_to_2D(point3D)
        point2D_shifted.x += length2D
        point2D_shifted = self.rectify(point2D_shifted)
        point3D_shifted_c = Point3D(self.Kinv @ point2D_shifted.H)
        point3D_c = Point3D(np.hstack((self.R, self.T)) @ point3D.H)  # Point3D expressed in camera coordinates system
        point3D_shifted_c = point3D_shifted_c*point3D_c.z/point3D_shifted_c.z
        length = np.linalg.norm(point3D_shifted_c - point3D_c, axis=0)
        return length


    def projects_in(self, point3D: Point3D) -> np.ndarray:
        """ Check wether point3D projects into the `Calib` object.
            Returns `True` where for points that projects in the image and `False` otherwise.
        """
        point2D = self.project_3D_to_2D(point3D)
        cond = np.stack((point2D.x >= 0, point2D.y >= 0, point2D.x <= self.width-1, point2D.y <= self.height-1))
        return np.all(cond, axis=0)

    def dist_to_border(self, point3D: Point3D) -> np.ndarray:
        """ Returns the distance to `Calib` object's borders in both dimensions,
            For each point given in point3D
        """
        point2D = self.project_3D_to_2D(point3D)
        distx = np.min(np.stack((self.width - point2D.x, point2D.x)), axis=0)
        disty = np.min(np.stack((self.height - point2D.y, point2D.y)), axis=0)
        return distx, disty

    def is_behind(self, point3D: Point3D) -> np.ndarray:
        """ Returns `True` where for points that are behind the camera and `False` otherwise.
        """
        n = Point3D(0,0,1) # normal to the camera plane in camera 3D coordinates system
        point3D_c = Point3D(np.hstack((self.R, self.T)) @ point3D.H)  # point3D expressed in camera 3D coordinates system
        return np.asarray((n.T @ point3D_c)[0] < 0)

    def __str__(self):
        return f"Calib object ({self.width}×{self.height})@({self.C.x: 1.6g},{self.C.y: 1.6g},{self.C.z: 1.6g})"


def line_plane_intersection(C: Point3D, d, P: Point3D, n) -> Point3D:
    """ Finds the intersection between a line and a plane.
        Args:
            C (Point3D): a point on the line
            d (np.ndarray): the direction-vector of the line
            P (Point3D): a Point3D on the plane
            n (np.ndarray): the normal vector of the plane
        Returns the Point3D at the intersection between the line and the plane.
    """
    d = d/np.linalg.norm(d, axis=0)
    dot = d.T @ n
    if np.any(np.abs(dot) < EPS):
        return None
    dist = ((P-C).T @ n) / dot  # Distance between plane z=Z and camera
    return Point3D(C + dist.T*d)


def point_line_distance(P, d, X) -> float:
    """ Finds the distance between a line and a point.
        Args:
            P (HomogeneousCoordinatesPoint): a point on the line
            d (np.ndarray): the line direction-vector
            X (HomogeneousCoordinatesPoint): a point.
        Returns the distance between the line and the point.
    """
    return np.linalg.norm(np.cross(d.flatten(), (P-X).flatten()))/np.linalg.norm(d)


def compute_rotate(width, height, angle, degrees=True):
    """ Computes rotation matrix and new width and height for a rotation of angle degrees of a widthxheight image.
    """
    assert degrees, "Angle in gradient is not implemented (yet)"
    # Convert the OpenCV 3x2 rotation matrix to 3x3
    R = np.eye(3)
    R[0:2,:] = cv2.getRotationMatrix2D((width/2, height/2), angle, 1.0)
    R2D = R[0:2,0:2]

    # Obtain the rotated coordinates of the image corners
    rotated_coords = [
        np.array([-width/2,  height/2]) @ R2D,
        np.array([ width/2,  height/2]) @ R2D,
        np.array([-width/2, -height/2]) @ R2D,
        np.array([ width/2, -height/2]) @ R2D
    ]

    # Find the size of the new image
    right_bound = max([pt[0] for pt in rotated_coords])
    left_bound = min([pt[0] for pt in rotated_coords])
    top_bound = max([pt[1] for pt in rotated_coords])
    bot_bound = min([pt[1] for pt in rotated_coords])

    new_width = int(abs(right_bound - left_bound))
    new_height = int(abs(top_bound - bot_bound))

    # We require a translation matrix to keep the image centred
    T = np.array([
        [1, 0, new_width/2 - width/2],
        [0, 1, new_height/2 - height/2],
        [0, 0, 1]
    ])
    return T@R, new_width, new_height


def rotate_image(image, angle):
    """ Rotates an image around its center by the given angle (in degrees).
        The returned image will be large enough to hold the entire new image, with a black background
    """
    height, width = image.shape[0:2]
    A, new_width, new_height = compute_rotate(width, height, angle)
    return cv2.warpAffine(image, A[0:2,:], (new_width, new_height), flags=cv2.INTER_LINEAR)

def parameters_to_affine_transform(angle: float, x_slice: slice, y_slice: slice,
    output_shape: tuple, do_flip: bool=False):
    """ Compute the affine transformation matrix that correspond to a
        - horizontal flip if `do_flip` is `True`, followed by a
        - rotation of `angle` degrees around output image center, followed by a
        - crop defined by `x_slice` and `y_slice`, followed by a
        - scale to recover `output_shape`.
    """
    # Rotation
    R = np.eye(3)
    center = ((x_slice.start + x_slice.stop)/2, (y_slice.start + y_slice.stop)/2)
    R[0:2,:] = cv2.getRotationMatrix2D(center, angle, 1.0)

    # Crop
    x0 = x_slice.start
    y0 = y_slice.start
    width = x_slice.stop - x_slice.start
    height = y_slice.stop - y_slice.start
    C = np.array([[1, 0,-x0], [0, 1,-y0], [0, 0, 1]])

    # Scale
    sx = output_shape[0]/width
    sy = output_shape[1]/height
    S = np.array([[sx, 0, 0], [0, sy, 0], [0, 0, 1]])

    # Random Flip
    assert not do_flip, "There is a bug with random flip"
    f = np.random.randint(0,2)*2-1 if do_flip else 1 # random sample in {-1,1}
    F = np.array([[f, 0, 0], [0, 1, 0], [0, 0, 1]])

    return S@C@R@F


def compute_rotation_matrix(point3D: Point3D, camera3D: Point3D):
    """ Computes the rotation matrix of a camera in `camera3D` pointing
        towards the point `point3D`. Both are expressed in word coordinates.
        The convention is that Z is pointing down.
        Credits: François Ledent
    """
    point3D = camera3D - point3D
    x, y, z = point3D.x, point3D.y, point3D.z
    d = np.sqrt(x**2 + y**2)
    D = np.sqrt(x**2 + y**2 + z**2)
    h = d / D
    l = z / D

    # camera basis `B` expressed in the world basis `O`
    _x = np.array([y / d, -x / d, 0])
    _y = np.array([- l * x / d, - l * y / d, h])
    _z = np.array([- h * x / d, - h * y / d, -l])
    B = np.stack((_x, _y, _z), axis=-1)

    # `O = R @ B` (where `O` := `np.identity(3)`)
    R = B.T # inv(B) == B.T since R is a rotation matrix
    return R


def compute_shear_rectification_matrix(calib: Calib, point2D: Point2D):
    """ Computes the transformation matrix that rectifies the image shear due to
        projection in the image plane, at the point `point2D`.
    """
    vector = Point2D(point2D.x - calib.K[0,2], point2D.y - calib.K[1,2])
    R1 = np.eye(3)
    R1[0:2,:] = cv2.getRotationMatrix2D(point2D.to_int_tuple(), np.arctan2(vector.y, vector.x)*180/np.pi, 1.0)
    scale = np.cos(np.arctan(np.linalg.norm(vector)/calib.K[0,0]))
    T1 = np.array([[1, 0, -point2D.x], [0, 1, -point2D.y], [0, 0, 1]])
    S = np.array([[scale, 0, 0], [0, 1, 0], [0, 0, 1]])
    R2 = np.eye(3)
    R2[0:2,:] = cv2.getRotationMatrix2D((0, 0), -np.arctan2(vector.y, vector.x)*180/np.pi, 1.0)
    T2 = np.array([[1, 0,  point2D.x], [0, 1,  point2D.y], [0, 0, 1]])
    return T2@R2@S@T1@R1



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    version="2.12.0",
    name='calib3d',
    author='Gabriel Van Zandycke',
    author_email="gabriel.vanzandycke@hotmail.com",
    url="https://github.com/ispgroupucl/calib3d",
    license="LGPL",
    python_requires='>=3.6',
    description="Python 3D calibration and homogenous coordinates computation library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    install_requires=[
        "numpy",
        "opencv-python"
    ],
    extras_requires={
        "tensorflow": ["tensorflow>=2.4"],
        "pycuda": ["pycuda"],
        "matplotlib": ["matplotlib"],
    }
)



from mako.template import Template
import numpy as np

from calib3d import Calib

class CudaCalib(Calib):
    datatypes = {
            'width':  ('__align__(8) int',    '',     8),
            'height': ('__align__(8) int',    '',     8),
            'T':      ('__align__(8) double', '[3]',  24),
            'K':      ('__align__(8) double', '[9]',  72),
            'kc':     ('__align__(8) double', '[5]',  40),
            'R':      ('__align__(8) double', '[9]' , 72),
            'C':      ('__align__(8) double', '[3]' , 24),
            'P':      ('__align__(8) double', '[12]', 96),
            'Pinv':   ('__align__(8) double', '[12]', 96),
            'Kinv':   ('__align__(8) double', '[9]',  72),
        }

    @classmethod
    def struct_str(cls):
        return Template("""
            struct calib_t
            {
                %for var in data:
                ${var[1]} ${var[0]}${var[2]};
                %endfor
            };
        """).render(data=[(name, dtype, size) for name, (dtype, size, _) in  cls.datatypes.items()])

    @classmethod
    def memsize(cls):
        return sum([size for attr, (_, _, size) in cls.datatypes.items()])

    @classmethod
    def from_calib(cls, calib):
        new_calib = cls.__new__(cls)
        new_calib.update(**calib.dict)
        return new_calib

    def memset(self, ptr, callback):
        # check function 'pycuda.tools.dtype_to_ctype'
        offset = 0
        for name, (_, _, size) in self.datatypes.items():
            value = self.__dict__[name]
            if isinstance(value, int):
                data = memoryview(np.int32(value))
            elif isinstance(value, float):
                data = memoryview(np.float32(value))
            else:
                data = memoryview(value)
            callback(int(ptr)+offset, data)
            offset += size
